/*------------------------------------------------------------------------*
 * Zen BASIC --- based on DDS's Tiny BASIC interpreter
 * Just enough to play Lunar Lander under Zen!
 *
 * $VER: ddsbasic v2.1
 * basic.c
 *
 * Untangled version of dds.c - originally by Diomidis Spinellis
 *                              entered in the IOCCC, http://www.ioccc.org/
 * Untangled, modified, Amigaized and improved by Adam Sampson
 *
 * Main mods:
 * o Works
 * o Error-checking in SAVE/OLD routines
 * o More-or-less readable
 * o Semicolon works at end of PRINT
 * o Modified to build successfully in GCC v2.6.1:
 *    use own stdio.h
 *    change f to a FILE * type
 *    get rid of duplicate malloc() and fopen() definitions
 *    change the puts()s to NOT print a \n - gcc's puts() does it anyway
 * o Fixed the line that I commented out in the previous version; NEXT now works! :)
 * modified for Hitec C and Zen
 * o Removed *lots* of VAXcentric assumptions
 * o Added non-interactive mode
 *
 * To do:
 * o Get rid of case dependance
 * o Fix that annoying line in G() with malloc(); the line array are not ptrs
 * o Retangle
 *
 *------------------------------------------------------------------------*/

/* This included instead of added bits from NorthC includes. */
#include "stdio.h"

/* Macro for search and replace routine */
#define	U(y,z) while(p=funQ(s,y)) *p++=z,*p=' '

/* number of lines const - don't s/rep */
#define R 500

typedef char *A;

int *C,E[R],funL[R],funM[R],funP[R],l=0,i=0,j=0;
char B[R],funF[2];
char *m[12*R],*p,*q,*x,*y,*z,*s,*d;
FILE *f;

/*--------------------- Function 'funQ' - equivalent to funQ() ----------*/
char *funQ(s,o)
     char *s,*o;
{
  for(x=s;*x;x++) {
    for(y=x,z=o;*z&&*y==*z;y++)
      z++;
    if(z>o&&!*z)
      return x;

  }
  return 0;
}

/*--------------------- Function 'main' - main loop -------------*/
int main(int argc, char **argv)
{ /* start of main() */
  int i, interactive=1;

  for(i=0;i<11*R;i++) m[i]=0; /* original was VAXcentric, */
  m[11*R]="E";

  if(argc==2) { /* invoked as basic <file>.bas */
    if(f=fopen(argv[1],"r")) {

      while(fgets(B,R,f)) {
	if (B[strlen(B)-1]=='\n') B[strlen(B)-1]='\0';
	G();
      }
      fclose(f);
      interactive=0;
    } else {
      printf("couldn't open '%s'\n",argv[1]);
    }
  } else {
    printf("Zen BASIC v2.1 by Diomidis Spinellis and Adam Sampson\n\n");
  }

  while(printf(interactive?"Ok\n":""),interactive?fgets(B,R,stdin):(B[0]='R')) {
    if (B[strlen(B)-1]=='\n') B[strlen(B)-1]='\0';
    switch(*B) {
    case 'R':  /* run */
      C=E;
      l=1;
      for(i=0;i<R;funP[i++]=0)
	;
      while(l) {		     /* while not at end of prog */
	while(!(s=m[l]))     /* find a non-empty line */
	  l++;
	if (!funQ(s,"\"")) {    /* search-replace if not a PRINT "str" */
	  U("<>",'#');
	  U("<=",'$');
	  U(">=",'!');
	}
	d=B;
	j=funF[0]=funF[1]=0;
	while(*funF=*s) {
	  if (*s=='"') j++;
	  if(j&1||!funQ(" \t",funF))
	    *d++=*s;
	  s++;
	}
	*d--=j=0;
	if(B[1]!='=')
	  switch(*B) {
	  case 'E': /* end */
	    l=-1;
	    break;
	  case 'R': /* rem */
	    if(B[2]!='M') (l=*--C);
	    break;
	  case 'I':
	    if (B[1]=='N') { /* input */
	      fgets(p=B,999,stdin);
	      funP[*d]=funS();
	    } else { /* if */
	      *(q=funQ(B,"TH"))=0;
	      p=B+2;
	      if (funS()) {
		p=q+4;
		l=funS()-1;
	      }
	    }
	    break;
	  case 'P':         /* print */
	    if (B[5]=='"') {
	      if (*d==';') {
		*--d=0;
		printf("%s",B+6);
	      } else {
		*d=0;
		printf("%s\n",B+6);
	      }
	    } else {
	      p=B+5;
	      printf("%d\n",funS());
	    }
	    break;
	  case 'G':         /* goto */
	    p=B+4;
	    if (B[2]=='S') { /* gosub */
	      (*C++=l,p++);
	    }
	    l=funS()-1;
	    break;
	  case 'F':         /* for */
	    *(q=funQ(B,"TO"))=0;
	    p=B+5;
	    funP[i=B[3]]=funS();
	    p=q+2;
	    funM[i]=funS();
	    funL[i]=l;
	    break;
	  case 'N':         /* next */
	    if (++funP[*d]<=funM[*d])
	      l=funL[*d]; /* restore li no */
	    else
	      funP[*d]=++funM[*d];
	  }
	else {
	  p=B+2;
	  funP[*B]=funS(); 
	}
	l++; /* on to the next line */
      };
      if (!interactive) goto howihatelabels;
      break;
    case 'L':                       /* list */
      for(i=0;i<11*R;i++)
	if (m[i]) printf("%d %s\n",i,m[i]);
      break;
    case 'N':                       /* new */
      for(i=0;i<11*R;i++) {
	if (m[i]) free(m[i]);
	m[i]=0;
      }
      break;
    case 'B':                       /* bye */
    howihatelabels:
      for(i=0;i<11*R;i++) if(m[i]) free(m[i]);
      return 0;
    case 'S':                      /* save */
      if(f=fopen(B+5,"w")) {
	for(i=0;i<11*R;i++)
	  if(m[i]) fprintf(f,"%d %s\n",i,m[i]);
	fclose(f);
      } else {
	printf("can't save\n");
      }
      break;
    case 'O':                      /* old */
      if(f=fopen(B+4,"r")) {
	while(fgets(B,R,f)) {
	  if (B[strlen(B)-1]=='\n') B[strlen(B)-1]='\0';
	  G();
	}
	fclose(f);
      } else {
	printf("can't load\n");
      }
      break;
    case 0:
    default:                        /* add line */
      G();
      break;
    } /* end of switch(B) */
  } /* end of while(q=0) */
  return 0;
} /* end of main */


/*------------- Function 'G' - allocate and add line -----------*/
int G()
{
  l=atoi(B);
  if(m[l]) free(m[l]);
  (p=funQ(B," "))?strcpy(m[l]=malloc(strlen(p)),p+1):(m[l]=0,0);
}

/*----------------------- Evaluate chain; funS,funJ,K,V,W ----------*/
int funS()
{
  int o=funJ();
  switch(*p++) {
  case '=': return o==funS();
  case '#': return o!=funS();
  default:	p--; return o;
  }
}

int funJ()
{
  int o=K();
  switch(*p++) {
  case '<': return o<funJ();
  case '>': return o>funJ();
  default:	p--; return o;
  }
}

int K()
{
  int o=V();
  switch(*p++) {
  case '$': return o<=K();
  case '!': return o>=K();
  default:	p--; return o;
  }
}

int V()
{
  int o=W();
  switch(*p++) {
  case '+': return o+V();
  case '-': return o-V();
  default:	p--; return o;
  }
}

int W()
{
  int o=funY();
  switch(*p++) {
  case '*': return o*W();
  case '/': return o/W();
  default:	p--; return o;
  }
}

/*---------- Function 'funY' - handle numbers, brackets and vars --------*/
int funY()
{
  int o;
  if (*p=='-') {
    p++;
    return -funY();
  } else {
    if (*p>='0'&&*p<='9') {
      return val();
    } else {
      if (*p=='(') {
	p++;
	o=funS();
	p++;
	return o;
      } else {
	return funP[*p++];
      }
    }
  }
}

/*---------------------- val - evaluate number ---------*/
int val()
{
  int a=0;
  while((*p>='0'&&*p<='9')&&strlen(p)>0) {
    a=(10*a)+(*p-'0');
    p++;
  }
  return a;
}
