#!/bin/sh

# Simplistic cc using cpm and the CP/M Hitec C binaries.
# azz@josstix.demon.co.uk

emu () {
    echo CP/M command: $*
    echo unix | cpm $*
    echo
}

set -- `getopt co: $*`
if [ $? != 0 ] ; then
    echo "Usage: hitec-cc.sh [-o OUTPUTFILENAME] [-c] { filename.{c,asm,o} | filename.o [ filename.o ... ] }"
    exit 2
fi
makebinaries=true
useoutputfilename=false
for i ; do
    case "$i" in
    -c)
	makebinaries=false
	shift
	;;
    -o)
	useoutputfilename=true
	outputfilename=$2
	shift
	shift
	;;
    --)
	shift
	break
	;;
    esac
done

runlinker=false
if [ $# -gt 1 -o $makebinaries = "true" ] ; then
    runlinker=true
fi
objects=""
for file in $* ; do
    type=$(echo $file | sed "s/.*\.\([^\.]*\)/\1/g")
    base=$(basename $file .$type)
    case "$type" in
    obj)
	$useoutputfilename && [ $# = 1 ] && mv -f $base.obj $outputfilename
	objects="$objects $file"
	;;
    lib)
	objects="$objects $file"
	;;
    c)
	# Neither cpp is perfect; the emulated one is better.
	sed "s/#include.*<\(.*\)>/#include \"\1\"/g" <$base.c >$base.cc
	emu cpp $base.cc $base.cpp
	#echo "Using normal cpp"
	#cpp -nostdinc -I. -Iinclude <$base.c | grep -v ^\# >$base.cpp
	emu p1 $base.cpp $base.p1
	emu cgen $base.p1 $base.gen
	emu optim $base.gen $base.as
	emu zas $base.as
	rm -f $base.cc $base.cpp $base.p1 $base.gen $base.as
	$useoutputfilename && [ $# = 1 -a $runlinker = "false" ] && mv -f $base.obj $outputfilename
	objects="$objects $base.obj"
	;;
    as)
	emu zas $base.as
	$useoutputfilename && [ $# = 1 ] && mv -f $base.obj $outputfilename
	objects="$objects $base.obj"
	;;
    esac
done

if [ $useoutputfilename = "false" ] ; then
    outputfilename=a.out
fi

if $runlinker ; then
    if $makebinaries ; then
	emu link -C0 -O$outputfilename initial.obj $objects libc.lib
    else
	emu link -R -O$outputfilename $objects
    fi
fi